@@ -180,7 +180,27 @@ int pureNullUnit(SEXP unit, int index) {
     else if (isUnitList(unit)) {
 	result = pureNullUnit(VECTOR_ELT(unit, index), 0);
     } else {  /* Just a plain unit */
-	result = unitUnit(unit, index) == L_NULL;
+	/* Special case:  if "grobwidth" or "grobheight" unit
+	 * and width/height(grob) is pure null
+	 */
+	if (unitUnit(unit, index) == L_GROBWIDTH) {
+	    SEXP fn, R_fcall;
+	    SEXP width;
+	    PROTECT(fn = findFun(install("width"), R_GlobalEnv));
+	    PROTECT(R_fcall = lang2(fn, unitData(unit, index)));
+	    PROTECT(width = eval(R_fcall, R_GlobalEnv));
+	    result = pureNullUnit(width, 0);
+	    UNPROTECT(3);
+	} else if (unitUnit(unit, index) == L_GROBHEIGHT) {
+	    SEXP fn, R_fcall;
+	    SEXP height;
+	    PROTECT(fn = findFun(install("height"), R_GlobalEnv));
+	    PROTECT(R_fcall = lang2(fn, unitData(unit, index)));
+	    PROTECT(height = eval(R_fcall, R_GlobalEnv));
+	    result = pureNullUnit(height, 0);
+	    UNPROTECT(3);
+	} else
+	    result = unitUnit(unit, index) == L_NULL;
     }
     return result;    
 }
@@ -196,16 +216,186 @@ int pureNullUnitArithmetic(SEXP unit, int index) {
     }
     else if (minFunc(unit) || maxFunc(unit) || sumFunc(unit)) {
 	int n = unitLength(arg1(unit));
-	int i;
+	int i = 0;
 	result = 1;
-	for (i=0; i<n; i++) 
+	while (result && i<n) {
 	    result = result && pureNullUnit(arg1(unit), i);
+	    i += 1;
+	}
     } 
     else 
 	error("Unimplemented unit function");
     return result;
 }
 
+/**************************
+ * Code for handling "grobwidth" units
+ **************************
+ */
+
+/* NOTE:  this code calls back to R code to perform 
+ * set.gpar operations, which will impact on grid global variables
+ * BUT that's ok(ish) because it subsequently calls back to R code to perform
+ * corresponding unset.gpar operations which will
+ * undo the changes to the grid globals
+ */
+
+double evaluateGrobWidthUnit(SEXP grob, 
+			     double vpfontsize, double vplineheight,
+			     double vpwidthCM, double vpheightCM,
+			     DevDesc *dd) 
+{
+    SEXP unit;
+    /* FIXME:  I probably want to create a new environment here
+     * rather than use the global environment (?) 
+     * Ditto in three eval()s below.
+     */
+    SEXP widthPreFn, widthFn, widthPostFn, R_fcall1, R_fcall2, R_fcall3;
+    SEXP getGParFn, R_getgparcall, gparname;
+    SEXP width, fontsize, lineheight;
+    LViewportContext vpc;
+    double resultINCHES, result;
+    PROTECT(widthPreFn = findFun(install("width.pre"), R_GlobalEnv));
+    PROTECT(widthFn = findFun(install("width"), R_GlobalEnv));
+    PROTECT(widthPostFn = findFun(install("width.post"), R_GlobalEnv));
+    PROTECT(getGParFn = findFun(install("get.gpar"), R_GlobalEnv));
+    /* Call width.pre(grob) 
+     */
+    PROTECT(R_fcall1 = lang2(widthPreFn, grob));
+    eval(R_fcall1, R_GlobalEnv);
+    /* Call width(grob)
+     * to get the unit representing the with
+     */
+    PROTECT(R_fcall2 = lang2(widthFn, grob));
+    PROTECT(width = eval(R_fcall2, R_GlobalEnv));
+    /* Call get.gpar() to get the current fontsize and lineheight settings
+     */
+    PROTECT(gparname = allocVector(STRSXP, 1));
+    SET_STRING_ELT(gparname, 0, mkChar("fontsize"));
+    PROTECT(R_getgparcall = lang2(getGParFn, gparname));
+    PROTECT(fontsize = eval(R_getgparcall, R_GlobalEnv));
+    SET_STRING_ELT(gparname, 0, mkChar("lineheight"));
+    R_getgparcall = lang2(getGParFn, gparname);
+    PROTECT(lineheight = eval(R_getgparcall, R_GlobalEnv));
+    /* Transform the width
+     * NOTE:  the width.pre function should NOT have done any 
+     * viewport pushing so the viewport context and sizeCM 
+     * should be unchanged
+     * NOTE:  We transform into INCHES so can produce final answer in terms
+     * of NPC for original context
+     */
+    /* Special case for "null" units
+     */
+    if (unitUnit(width, 0) == L_NULL) {
+	result = evaluateNullUnit(unitValue(width, 0));
+    } else {
+	/* These are the current fontsize and lineheight.
+	 * They are adequate ONLY because of the strict
+	 * restrictions placed on what sort of units can
+	 * be used in a width.details method.
+	 * If those restrictions are lifted, all bets are off ...
+	 */
+	vpc.fontsize = vpfontsize;
+	vpc.lineheight = vplineheight;
+	resultINCHES = transformWidthtoINCHES(width, 0, vpc,
+					      REAL(fontsize)[0], 
+					      REAL(lineheight)[0],
+					      /* These are the current
+					       * viewport widthCM and
+					       * heightCM.
+					       * They are adequate ONLY
+					       * because of the strict
+					       * restrictions placed on
+					       * what sort of units can
+					       * be used in a width.details
+					       * method.
+					       * If those restrictions are
+					       * lifted, all bets are off ...
+					       */
+					      vpwidthCM, vpheightCM,
+					      dd);
+	result = resultINCHES/(vpwidthCM/2.54);
+    }
+    /* Call width.post(grob)
+     */
+    PROTECT(R_fcall3 = lang2(widthPostFn, grob));
+    eval(R_fcall3, R_GlobalEnv);
+    UNPROTECT(12);
+    /* Return the transformed width
+     */
+    return result;
+}
+
+/* See evaluateGrobWidthUnit for detailed comments
+ */
+double evaluateGrobHeightUnit(SEXP grob, 
+			     double vpfontsize, double vplineheight,
+			     double vpheightCM, double vpwidthCM,
+			     DevDesc *dd) 
+{
+    SEXP unit;
+    /* FIXME:  I probably want to create a new environment here
+     * rather than use the global environment (?) 
+     * Ditto in three eval()s below.
+     */
+    SEXP heightPreFn, heightFn, heightPostFn, R_fcall1, R_fcall2, R_fcall3;
+    SEXP getGParFn, R_getgparcall, gparname;
+    SEXP height, fontsize, lineheight;
+    LViewportContext vpc;
+    double resultINCHES, result;
+    PROTECT(heightPreFn = findFun(install("height.pre"), R_GlobalEnv));
+    PROTECT(heightFn = findFun(install("height"), R_GlobalEnv));
+    PROTECT(heightPostFn = findFun(install("height.post"), R_GlobalEnv));
+    PROTECT(getGParFn = findFun(install("get.gpar"), R_GlobalEnv));
+    /* Call height.pre(grob) 
+     */
+    PROTECT(R_fcall1 = lang2(heightPreFn, grob));
+    eval(R_fcall1, R_GlobalEnv);
+    /* Call height(grob)
+     * to get the unit representing the with
+     */
+    PROTECT(R_fcall2 = lang2(heightFn, grob));
+    PROTECT(height = eval(R_fcall2, R_GlobalEnv));
+    /* Call get.gpar() to get the current fontsize and lineheight settings
+     */
+    PROTECT(gparname = allocVector(STRSXP, 1));
+    SET_STRING_ELT(gparname, 0, mkChar("fontsize"));
+    PROTECT(R_getgparcall = lang2(getGParFn, gparname));
+    PROTECT(fontsize = eval(R_getgparcall, R_GlobalEnv));
+    SET_STRING_ELT(gparname, 0, mkChar("lineheight"));
+    R_getgparcall = lang2(getGParFn, gparname);
+    PROTECT(lineheight = eval(R_getgparcall, R_GlobalEnv));
+    /* Transform the height
+     * NOTE:  the height.pre function should NOT have done any 
+     * viewport pushing so the viewport context and sizeCM 
+     * should be unchanged
+     * NOTE:  We transform into INCHES so can produce final answer in terms
+     * of NPC for original context
+     */
+    /* Special case for "null" units
+     */
+    if (unitUnit(height, 0) == L_NULL) {
+	result = evaluateNullUnit(unitValue(height, 0));
+    } else {
+	vpc.fontsize = vpfontsize;
+	vpc.lineheight = vplineheight;
+	resultINCHES = transformHeighttoINCHES(height, 0, vpc,
+					      REAL(fontsize)[0], 
+					      REAL(lineheight)[0],
+					      vpwidthCM, vpheightCM,
+					      dd);
+	result = resultINCHES/(vpheightCM/2.54);
+    }
+    /* Call height.post(grob)
+     */
+    PROTECT(R_fcall3 = lang2(heightPostFn, grob));
+    eval(R_fcall3, R_GlobalEnv);
+    UNPROTECT(12);
+    /* Return the transformed height
+     */
+    return result;
+}
+
 /**************************
  * TRANSFORMATIONS
  **************************
@@ -216,6 +406,7 @@ int pureNullUnitArithmetic(SEXP unit, int index) {
 double transform(double value, int unit, SEXP data,
 		 double scalemin, double scalemax,
 		 double fontSize, double lineHeight,
+		 double grobfontSize, double groblineHeight,
 		 double thisCM, double otherCM,
 		 DevDesc *dd)
 {
@@ -240,9 +431,15 @@ double transform(double value, int unit, SEXP data,
     case L_CHAR:
 	result = result*fontSize/(72*thisCM/2.54);
 	break;	
+    case L_MYCHAR:
+	result = result*grobfontSize/(72*thisCM/2.54);
+	break;	
     case L_LINES:
 	result = result*fontSize*lineHeight/(72*thisCM/2.54);
 	break;
+    case L_MYLINES:
+	result = result*fontSize*groblineHeight/(72*thisCM/2.54);
+	break;
     case L_SNPC:        
 	if (thisCM <= otherCM)
 	    result = result;
@@ -280,6 +477,14 @@ double transform(double value, int unit, SEXP data,
 	result = result*GStrHeight(CHAR(STRING_ELT(data, 0)), INCHES, dd)*
 	    2.54/thisCM;
 	break;
+    case L_GROBWIDTH:
+	result = value*evaluateGrobWidthUnit(data, fontSize, lineHeight,
+					     thisCM, otherCM, dd);
+	break;
+    case L_GROBHEIGHT:
+	result = value*evaluateGrobHeightUnit(data, fontSize, lineHeight,
+					      thisCM, otherCM, dd);
+	break;
     case L_NULL:
 	result = evaluateNullUnit(result);
 	break;
@@ -293,6 +498,7 @@ double transform(double value, int unit, SEXP data,
 double transformLocation(double location, int unit, SEXP data,
 			 double scalemin, double scalemax,
 			 double fontSize, double lineHeight,
+			 double grobfontSize, double groblineHeight,
 			 double thisCM, double otherCM,
 			 DevDesc *dd)
 {
@@ -303,18 +509,21 @@ double transformLocation(double location, int unit, SEXP data,
 	break;
     default:
 	result = transform(location, unit, data, scalemin, scalemax,
-			   fontSize, lineHeight, thisCM, otherCM, dd);
+			   fontSize, lineHeight, grobfontSize, groblineHeight, 
+			   thisCM, otherCM, dd);
     }
     return result;
 }
 
 double transformXArithmetic(SEXP x, int index,
 			    LViewportContext vpc,
+			    double fontsize, double lineheight,
 			    double widthCM, double heightCM,
 			    DevDesc *dd);
 
 double transformX(SEXP x, int index,
 		  LViewportContext vpc,
+		  double fontsize, double lineheight,
 		  double widthCM, double heightCM,
 		  DevDesc *dd)
 {
@@ -322,10 +531,13 @@ double transformX(SEXP x, int index,
     int unit;
     SEXP data;
     if (isUnitArithmetic(x)) 
-	result = transformXArithmetic(x, index, vpc, widthCM, heightCM, dd);
+	result = transformXArithmetic(x, index, vpc, 
+				      fontsize, lineheight,
+				      widthCM, heightCM, dd);
     else if (isUnitList(x)) {
 	int n = unitLength(x);
 	result = transformX(VECTOR_ELT(x, index % n), 0, vpc, 
+			    fontsize, lineheight,
 			    widthCM, heightCM, dd);
     } else {  /* Just a plain unit */
 	L_nullArithmeticMode = L_plain;
@@ -335,6 +547,8 @@ double transformX(SEXP x, int index,
 	result = transformLocation(result, unit, data, 
 				   vpc.xscalemin, vpc.xscalemax,
 				   vpc.fontsize, vpc.lineheight,
+				   fontsize, 
+				   lineheight,
 				   widthCM, heightCM, dd);
 	switch (vpc.origin) {
 	case L_BOTTOMLEFT:         
@@ -349,13 +563,27 @@ double transformX(SEXP x, int index,
     return result;
 }
 
+double transformXtoNative(SEXP x, int index,
+			  LViewportContext vpc,
+			  double fontsize, double lineheight,
+			  double widthCM, double heightCM,
+			  DevDesc *dd)
+{
+    return vpc.xscalemin + 
+	transformX(x, index, vpc, fontsize, lineheight, 
+		   widthCM, heightCM, dd)*
+	(vpc.xscalemax - vpc.xscalemin);
+}
+
 double transformYArithmetic(SEXP y, int index,
 			    LViewportContext vpc,
+			    double fontsize, double lineheight,
 			    double widthCM, double heightCM,
 			    DevDesc *dd);
 
 double transformY(SEXP y, int index, 
 		  LViewportContext vpc,
+		  double fontsize, double lineheight,
 		  double widthCM, double heightCM,
 		  DevDesc *dd)
 {
@@ -364,10 +592,12 @@ double transformY(SEXP y, int index,
     SEXP data;
     if (isUnitArithmetic(y))
 	result = transformYArithmetic(y, index, vpc,
+				      fontsize, lineheight,
 				      widthCM, heightCM, dd);
     else if (isUnitList(y)) {
 	int n = unitLength(y);
 	result = transformY(VECTOR_ELT(y, index % n), 0, vpc,
+			    fontsize, lineheight,
 			    widthCM, heightCM, dd);
     } else { /* Just a unit object */
 	L_nullArithmeticMode = L_plain;
@@ -377,6 +607,8 @@ double transformY(SEXP y, int index,
 	result = transformLocation(result, unit, data, 
 				   vpc.yscalemin, vpc.yscalemax,
 				   vpc.fontsize, vpc.lineheight,
+				   fontsize, 
+				   lineheight,
 				   heightCM, widthCM, dd);
 	switch (vpc.origin) {
 	case L_BOTTOMLEFT:
@@ -391,9 +623,22 @@ double transformY(SEXP y, int index,
     return result;
 }
 
+double transformYtoNative(SEXP y, int index,
+			  LViewportContext vpc,
+			  double fontsize, double lineheight,
+			  double widthCM, double heightCM,
+			  DevDesc *dd)
+{
+    return vpc.yscalemin + 
+	transformY(y, index, vpc, fontsize, lineheight, 
+		   widthCM, heightCM, dd)*
+	(vpc.yscalemax - vpc.yscalemin);
+}
+
 double transformDimension(double dim, int unit, SEXP data,
 			  double scalemin, double scalemax,
 			  double fontSize, double lineHeight,
+			  double grobfontSize, double groblineHeight,
 			  double thisCM, double otherCM,
 			  DevDesc *dd)
 {
@@ -404,18 +649,21 @@ double transformDimension(double dim, int unit, SEXP data,
 	break;
     default:
 	result = transform(dim, unit, data, scalemin, scalemax,
-			   fontSize, lineHeight, thisCM, otherCM, dd);
+			   fontSize, lineHeight, grobfontSize, groblineHeight, 
+			   thisCM, otherCM, dd);
     }
     return result;
 }
 
 double transformWidthArithmetic(SEXP width, int index,
 				LViewportContext vpc,
+				double fontsize, double lineheight,
 				double widthCM, double heightCM,
 				DevDesc *dd);
 
 double transformWidth(SEXP width, int index, 
 		      LViewportContext vpc,
+		      double fontsize, double lineheight,
 		      double widthCM, double heightCM,
 		      DevDesc *dd)
 {
@@ -424,10 +672,12 @@ double transformWidth(SEXP width, int index,
     SEXP data;
     if (isUnitArithmetic(width))
 	result = transformWidthArithmetic(width, index, vpc,
+					  fontsize, lineheight,
 					  widthCM, heightCM, dd);
     else if (isUnitList(width)) {
 	int n = unitLength(width);
 	result = transformWidth(VECTOR_ELT(width, index % n), 0, vpc,
+				fontsize, lineheight,
 				widthCM, heightCM, dd);
     } else { /* Just a unit object */
 	L_nullArithmeticMode = L_plain;
@@ -437,6 +687,8 @@ double transformWidth(SEXP width, int index,
 	result = transformDimension(result, unit, data, 
 				    vpc.xscalemin, vpc.xscalemax,
 				    vpc.fontsize, vpc.lineheight,
+				    fontsize, 
+				    lineheight,
 				    widthCM, heightCM, dd);
 	switch (vpc.origin) {
 	case L_BOTTOMLEFT:
@@ -451,13 +703,26 @@ double transformWidth(SEXP width, int index,
     return result;
 }
 
+double transformWidthtoNative(SEXP width, int index,
+			  LViewportContext vpc,
+			  double fontsize, double lineheight,
+			  double widthCM, double heightCM,
+			  DevDesc *dd)
+{
+    return transformWidth(width, index, vpc, fontsize, lineheight, 
+			  widthCM, heightCM, dd)*
+	(vpc.xscalemax - vpc.xscalemin);
+}
+
 double transformHeightArithmetic(SEXP height, int index,
 				 LViewportContext vpc,
+				 double fontsize, double lineheight,
 				 double widthCM, double heightCM,
 				 DevDesc *dd);
 
 double transformHeight(SEXP height, int index,
 		       LViewportContext vpc,
+		       double fontsize, double lineheight,
 		       double widthCM, double heightCM,
 		       DevDesc *dd)
 {
@@ -466,10 +731,12 @@ double transformHeight(SEXP height, int index,
     SEXP data;
     if (isUnitArithmetic(height))
 	result = transformHeightArithmetic(height, index, vpc,
+					   fontsize, lineheight,
 					   widthCM, heightCM, dd);
     else if (isUnitList(height)) {
 	int n = unitLength(height);
 	result = transformHeight(VECTOR_ELT(height, index % n), 0, vpc,
+				 fontsize, lineheight,
 				 widthCM, heightCM, dd);
     } else { /* Just a unit object */
 	L_nullArithmeticMode = L_plain;
@@ -479,6 +746,8 @@ double transformHeight(SEXP height, int index,
 	result = transformDimension(result, unit, data, 
 				    vpc.yscalemin, vpc.yscalemax,
 				    vpc.fontsize, vpc.lineheight,
+				    fontsize, 
+				    lineheight,
 				    heightCM, widthCM, dd);
 	switch (vpc.origin) {
 	case L_BOTTOMLEFT:
@@ -493,8 +762,20 @@ double transformHeight(SEXP height, int index,
     return result;
 }
 
+double transformHeighttoNative(SEXP height, int index,
+			  LViewportContext vpc,
+			  double fontsize, double lineheight,
+			  double widthCM, double heightCM,
+			  DevDesc *dd)
+{
+    return transformHeight(height, index, vpc, fontsize, lineheight, 
+			  widthCM, heightCM, dd)*
+	(vpc.yscalemax - vpc.yscalemin);
+}
+
 double transformXArithmetic(SEXP x, int index,
 			    LViewportContext vpc,
+			    double fontsize, double lineheight,
 			    double widthCM, double heightCM,
 			    DevDesc *dd)
 {
@@ -503,21 +784,26 @@ double transformXArithmetic(SEXP x, int index,
     if (addOp(x)) {
 	L_nullArithmeticMode = L_adding;
 	result = transformX(arg1(x), index, vpc,
+			    fontsize, lineheight,
 			    widthCM, heightCM, dd) +
 	    transformX(arg2(x), index, vpc,
+		       fontsize, lineheight,
 		       widthCM, heightCM, dd);
     }
     else if (minusOp(x)) {
 	L_nullArithmeticMode = L_subtracting;
 	result = transformX(arg1(x), index, vpc,
+			    fontsize, lineheight,
 			    widthCM, heightCM, dd) -
 	    transformX(arg2(x), index, vpc,
+		       fontsize, lineheight,
 		       widthCM, heightCM, dd);
     }
     else if (timesOp(x)) {
 	L_nullArithmeticMode = L_multiplying;
 	result = REAL(arg1(x))[0] *
 	    transformX(arg2(x), index, vpc,
+		       fontsize, lineheight,
 		       widthCM, heightCM, dd);
     }
     else if (minFunc(x)) {
@@ -525,9 +811,11 @@ double transformXArithmetic(SEXP x, int index,
 	double temp = DBL_MAX;
 	L_nullArithmeticMode = L_minimising;
 	result = transformX(arg1(x), 0, vpc,
+			    fontsize, lineheight,
 			    widthCM, heightCM, dd);
 	for (i=1; i<n; i++) {
 	    temp = transformX(arg1(x), i, vpc,
+			      fontsize, lineheight,
 			      widthCM, heightCM, dd);
 	    if (temp < result)
 		result = temp;
@@ -538,9 +826,11 @@ double transformXArithmetic(SEXP x, int index,
 	double temp = DBL_MIN;
 	L_nullArithmeticMode = L_maximising;
 	result = transformX(arg1(x), 0, vpc,
+			    fontsize, lineheight,
 			    widthCM, heightCM, dd);
 	for (i=1; i<n; i++) {
 	    temp = transformX(arg1(x), i, vpc,
+			      fontsize, lineheight,
 			      widthCM, heightCM, dd);
 	    if (temp > result)
 		result = temp;
@@ -552,6 +842,7 @@ double transformXArithmetic(SEXP x, int index,
 	L_nullArithmeticMode = L_summing;
 	for (i=0; i<n; i++) {
 	    result += transformX(arg1(x), i, vpc,
+				 fontsize, lineheight,
 				 widthCM, heightCM, dd);
 	}
     }
@@ -562,6 +853,7 @@ double transformXArithmetic(SEXP x, int index,
 
 double transformYArithmetic(SEXP y, int index,
 			    LViewportContext vpc,
+			    double fontsize, double lineheight,
 			    double widthCM, double heightCM,
 			    DevDesc *dd)
 {
@@ -570,21 +862,26 @@ double transformYArithmetic(SEXP y, int index,
     if (addOp(y)) {
 	L_nullArithmeticMode = L_adding;
 	result = transformY(arg1(y), index, vpc,
+			    fontsize, lineheight,
 			    widthCM, heightCM, dd) +
 	    transformY(arg2(y), index, vpc,
+		       fontsize, lineheight,
 		       widthCM, heightCM, dd);
     }
     else if (minusOp(y)) {
 	L_nullArithmeticMode = L_subtracting;
 	result = transformY(arg1(y), index, vpc,
+			    fontsize, lineheight,
 			    widthCM, heightCM, dd) -
 	    transformY(arg2(y), index, vpc,
+		       fontsize, lineheight,
 		       widthCM, heightCM, dd);
     }
     else if (timesOp(y)) {
 	L_nullArithmeticMode = L_multiplying;
 	result = REAL(arg1(y))[0] *
 	    transformY(arg2(y), index, vpc,
+		       fontsize, lineheight,
 		       widthCM, heightCM, dd);
     }
     else if (minFunc(y)) {
@@ -592,9 +889,11 @@ double transformYArithmetic(SEXP y, int index,
 	double temp = DBL_MAX;
 	L_nullArithmeticMode = L_minimising;
 	result = transformY(arg1(y), 0, vpc,
+			    fontsize, lineheight,
 			    widthCM, heightCM, dd);
 	for (i=1; i<n; i++) {
 	    temp = transformY(arg1(y), i, vpc,
+			      fontsize, lineheight,
 			      widthCM, heightCM, dd);
 	    if (temp < result)
 		result = temp;
@@ -605,9 +904,11 @@ double transformYArithmetic(SEXP y, int index,
 	double temp = DBL_MIN;
 	L_nullArithmeticMode = L_maximising;
 	result = transformY(arg1(y), 0, vpc,
+			    fontsize, lineheight,
 			    widthCM, heightCM, dd);
 	for (i=1; i<n; i++) {
 	    temp = transformY(arg1(y), i, vpc,
+			      fontsize, lineheight,
 			      widthCM, heightCM, dd);
 	    if (temp > result)
 		result = temp;
@@ -619,6 +920,7 @@ double transformYArithmetic(SEXP y, int index,
 	result = 0.0;
 	for (i=0; i<n; i++) {
 	    result += transformY(arg1(y), i, vpc,
+				 fontsize, lineheight,
 				 widthCM, heightCM, dd);
 	}
     }
@@ -629,6 +931,7 @@ double transformYArithmetic(SEXP y, int index,
 
 double transformWidthArithmetic(SEXP width, int index,
 				LViewportContext vpc,
+				double fontsize, double lineheight,
 				double widthCM, double heightCM,
 				DevDesc *dd)
 {
@@ -637,21 +940,26 @@ double transformWidthArithmetic(SEXP width, int index,
     if (addOp(width)) {
 	L_nullArithmeticMode = L_adding;
 	result = transformWidth(arg1(width), index, vpc,
+				fontsize, lineheight,
 				widthCM, heightCM, dd) +
 	    transformWidth(arg2(width), index, vpc,
+			   fontsize, lineheight,
 			   widthCM, heightCM, dd);
     }
     else if (minusOp(width)) {
 	L_nullArithmeticMode = L_subtracting;
 	result = transformWidth(arg1(width), index, vpc,
+				fontsize, lineheight,
 				widthCM, heightCM, dd) -
 	    transformWidth(arg2(width), index, vpc,
+			   fontsize, lineheight,
 			   widthCM, heightCM, dd);
     }
     else if (timesOp(width)) {
 	L_nullArithmeticMode = L_multiplying;
 	result = REAL(arg1(width))[0] *
 	    transformWidth(arg2(width), index, vpc,
+			   fontsize, lineheight,
 			   widthCM, heightCM, dd);
     }
     else if (minFunc(width)) {
@@ -659,9 +967,11 @@ double transformWidthArithmetic(SEXP width, int index,
 	double temp = DBL_MAX;
 	L_nullArithmeticMode = L_minimising;
 	result = transformWidth(arg1(width), 0, vpc,
+				fontsize, lineheight,
 				widthCM, heightCM, dd);
 	for (i=1; i<n; i++) {
 	    temp = transformWidth(arg1(width), i, vpc,
+				  fontsize, lineheight,
 				  widthCM, heightCM, dd);
 	    if (temp < result)
 		result = temp;
@@ -672,9 +982,11 @@ double transformWidthArithmetic(SEXP width, int index,
 	double temp = DBL_MIN;
 	L_nullArithmeticMode = L_maximising;
 	result = transformWidth(arg1(width), 0, vpc,
+				fontsize, lineheight,
 				widthCM, heightCM, dd);
 	for (i=1; i<n; i++) {
 	    temp = transformWidth(arg1(width), i, vpc,
+				  fontsize, lineheight,
 				  widthCM, heightCM, dd);
 	    if (temp > result)
 		result = temp;
@@ -686,6 +998,7 @@ double transformWidthArithmetic(SEXP width, int index,
 	L_nullArithmeticMode = L_summing;
 	for (i=0; i<n; i++) {
 	    result += transformWidth(arg1(width), i, vpc,
+				     fontsize, lineheight,
 				     widthCM, heightCM, dd);
 	}
     }
@@ -696,6 +1009,7 @@ double transformWidthArithmetic(SEXP width, int index,
 
 double transformHeightArithmetic(SEXP height, int index,
 				 LViewportContext vpc,
+				 double fontsize, double lineheight,
 				 double widthCM, double heightCM,
 				 DevDesc *dd)
 {
@@ -704,21 +1018,26 @@ double transformHeightArithmetic(SEXP height, int index,
     if (addOp(height)) {
 	L_nullArithmeticMode = L_adding;
 	result = transformHeight(arg1(height), index, vpc,
+				 fontsize, lineheight,
 				 widthCM, heightCM, dd) +
 	    transformHeight(arg2(height), index, vpc,
+			    fontsize, lineheight,
 			    widthCM, heightCM, dd);
     }
     else if (minusOp(height)) {
 	L_nullArithmeticMode = L_subtracting;
 	result = transformHeight(arg1(height), index, vpc,
+				 fontsize, lineheight,
 				 widthCM, heightCM, dd) -
 	    transformHeight(arg2(height), index, vpc,
+			    fontsize, lineheight,
 			    widthCM, heightCM, dd);
     }
     else if (timesOp(height)) {
 	L_nullArithmeticMode = L_multiplying;
 	result = REAL(arg1(height))[0] *
 	    transformHeight(arg2(height), index, vpc,
+			    fontsize, lineheight,
 			    widthCM, heightCM, dd);
     }
     else if (minFunc(height)) {
@@ -726,9 +1045,11 @@ double transformHeightArithmetic(SEXP height, int index,
 	double temp = DBL_MAX;
 	L_nullArithmeticMode = L_minimising;
 	result = transformHeight(arg1(height), 0, vpc,
+				 fontsize, lineheight,
 				 widthCM, heightCM, dd);
 	for (i=1; i<n; i++) {
 	    temp = transformHeight(arg1(height), i, vpc,
+				   fontsize, lineheight,
 				   widthCM, heightCM, dd);
 	    if (temp < result)
 		result = temp;
@@ -739,9 +1060,11 @@ double transformHeightArithmetic(SEXP height, int index,
 	double temp = DBL_MIN;
 	L_nullArithmeticMode = L_maximising;
 	result = transformHeight(arg1(height), 0, vpc,
+				 fontsize, lineheight,
 				 widthCM, heightCM, dd);
 	for (i=1; i<n; i++) {
 	    temp = transformHeight(arg1(height), i, vpc,
+				   fontsize, lineheight,
 				   widthCM, heightCM, dd);
 	    if (temp > result)
 		result = temp;
@@ -753,6 +1076,7 @@ double transformHeightArithmetic(SEXP height, int index,
 	result = 0.0;
 	for (i=0; i<n; i++) {
 	    result += transformHeight(arg1(height), i, vpc,
+				      fontsize, lineheight,
 				      widthCM, heightCM, dd);
 	}
     }
@@ -774,24 +1098,35 @@ double transformHeightArithmetic(SEXP height, int index,
  * In other words, the reason for the apparent inefficiency here
  * is historical.
  */
+
+/* The difference between transform*toINCHES and transformLocn/Dimn 
+ * is that the former are just converting from one coordinate system
+ * to INCHES;  the latter are converting from INCHES relative to
+ * the parent to INCHES relative to the device.
+ */
 double transformXtoINCHES(SEXP x, int index, 
 			  LViewportContext vpc,
+			  double fontsize, double lineheight,
 			  double widthCM, double heightCM,
 			  DevDesc *dd)
 {
-    return transformX(x, index, vpc, widthCM, heightCM, dd)*widthCM/2.54;
+    return transformX(x, index, vpc, fontsize, lineheight,
+		      widthCM, heightCM, dd)*widthCM/2.54;
 }
 
 double transformYtoINCHES(SEXP y, int index, 
 			  LViewportContext vpc,
+			  double fontsize, double lineheight,
 			  double widthCM, double heightCM,
 			  DevDesc *dd)
 {
-    return transformY(y, index, vpc, widthCM, heightCM, dd)*heightCM/2.54;
+    return transformY(y, index, vpc, fontsize, lineheight,
+		      widthCM, heightCM, dd)*heightCM/2.54;
 }
 
 void transformLocn(SEXP x, SEXP y, int index, 
 		   LViewportContext vpc,
+		   double fontsize, double lineheight,
 		   double widthCM, double heightCM,
 		   DevDesc *dd,
 		   LTransform t,
@@ -801,8 +1136,10 @@ void transformLocn(SEXP x, SEXP y, int index,
     /* x and y are unit objects (i.e., values in any old coordinate
      * system) so the first step is to convert them both to CM
      */
-    *xx = transformXtoINCHES(x, index, vpc, widthCM, heightCM, dd);
-    *yy = transformYtoINCHES(y, index, vpc, widthCM, heightCM, dd);
+    *xx = transformXtoINCHES(x, index, vpc, fontsize, lineheight,
+			     widthCM, heightCM, dd);
+    *yy = transformYtoINCHES(y, index, vpc, fontsize, lineheight,
+			     widthCM, heightCM, dd);
     location(*xx, *yy, lin);
     trans(lin, t, lout);
     *xx = locationX(lout);
@@ -811,22 +1148,27 @@ void transformLocn(SEXP x, SEXP y, int index,
 
 double transformWidthtoINCHES(SEXP w, int index,
 			      LViewportContext vpc,
+			      double fontsize, double lineheight,
 			      double widthCM, double heightCM,
 			      DevDesc *dd)
 {
-    return transformWidth(w, index, vpc, widthCM, heightCM, dd)*widthCM/2.54;
+    return transformWidth(w, index, vpc, fontsize, lineheight,
+			  widthCM, heightCM, dd)*widthCM/2.54;
 }
 
 double transformHeighttoINCHES(SEXP h, int index,
 			       LViewportContext vpc,
+			       double fontsize, double lineheight,
 			       double widthCM, double heightCM,
 			       DevDesc *dd)
 {
-    return transformHeight(h, index, vpc, widthCM, heightCM, dd)*heightCM/2.54;
+    return transformHeight(h, index, vpc, fontsize, lineheight,
+			   widthCM, heightCM, dd)*heightCM/2.54;
 }
 
 void transformDimn(SEXP w, SEXP h, int index, 
 		   LViewportContext vpc,
+		   double fontsize, double lineheight,
 		   double widthCM, double heightCM,
 		   DevDesc *dd,
 		   double rotationAngle,
@@ -834,9 +1176,9 @@ void transformDimn(SEXP w, SEXP h, int index,
 {
     LLocation din, dout;
     LTransform r;
-    *ww = transformWidthtoINCHES(w, index, vpc,
+    *ww = transformWidthtoINCHES(w, index, vpc, fontsize, lineheight,
 				 widthCM, heightCM, dd);
-    *hh = transformHeighttoINCHES(h, index, vpc,
+    *hh = transformHeighttoINCHES(h, index, vpc, fontsize, lineheight,
 				  widthCM, heightCM, dd);
     location(*ww, *hh, din);
     rotation(rotationAngle, r);
@@ -846,5 +1188,75 @@ void transformDimn(SEXP w, SEXP h, int index,
 }
 
 
+/* Attempt to make validating units faster
+ */
+typedef struct {
+    char *name;
+    int code;
+} UnitTab;
+
+/* NOTE this table must match the order in grid.h
+ */
+static UnitTab UnitTable[] = {
+    { "npc",            0 },
+    { "cm",             1 },
+    { "inches",         2 },
+    { "lines",          3 },
+    { "native",         4 },
+    { "null",           5 },
+    { "snpc",           6 },
+    { "mm",             7 },
+    { "points",         8 },
+    { "picas",          9 },
+    { "bigpts",        10 },
+    { "dida",          11 },
+    { "cicero",        12 },
+    { "scaledpts",     13 },
+    { "strwidth",      14 },
+    { "strheight",     15 },
+
+    { "char",          18 },
+    { "grobwidth",     19 },
+    { "grobheight",    20 },
+    { "mylines",       21 },
+    { "mychar",        22 },
+
+    { NULL,            -1 }
+};
+
+int convertUnit(SEXP unit, int index) 
+{
+    int i = 0;
+    int result = 0;
+    int found = 0;
+    while (result >= 0 && !found) {
+	if (UnitTable[i].name == NULL) 
+	    result = -1;
+	else {
+	    found = !strcmp(CHAR(STRING_ELT(unit, index)), UnitTable[i].name);
+	    if (found) 
+		result = UnitTable[i].code;
+	}
+	i += 1;
+    }
+    if (result < 0)
+	error("Invalid unit");
+    return result;
+}
+	    
+SEXP validUnits(SEXP units) 
+{
+    int i;
+    int n = LENGTH(units);
+    SEXP answer;
+    PROTECT(answer = allocVector(INTSXP, n));
+    for (i = 0; i<n; i++) 
+	INTEGER(answer)[i] = convertUnit(units, i);
+    UNPROTECT(1);
+    return answer;
+}
+    
+
+
 
 
